#  SPDX-FileCopyrightText: 2023 easyCrystallography contributors <crystallography@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2022-2023  Contributors to the easyCore project <https://github.com/easyScience/easyCrystallography>

from __future__ import annotations
__author__ = 'github.com/wardsimon'
__version__ = '0.0.1'

import textwrap

from collections.abc import MutableMapping
from gemmi import cif
from math import floor, log10
from numbers import Number
from typing import List, TYPE_CHECKING, Any, Optional, Dict, Union

from easyCore import np
from easyCore.Utils.io.template import BaseEncoderDecoder
from easyCore.Utils.io.dict import DictSerializer, DataDictSerializer

if TYPE_CHECKING:
    from easyCore.Utils.typing import B, V, BV

_MAX_LEN = 160
_SEP = '.'

def _flatten_dict_gen(d, parent_key, sep):
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            yield from flatten_dict(v, new_key, sep=sep).items()
        elif isinstance(v, list):
            yield new_key, [value if not isinstance(value, MutableMapping) else flatten_dict(value, '', sep) for value in v ]
        else:
            yield new_key, v


def flatten_dict(d: MutableMapping, parent_key: str = '', sep: str = '.') -> dict:
    return dict(_flatten_dict_gen(d, parent_key, sep))


def _unflatten_dict(key, value, out, sep) -> dict:
    key, *rest = key.split(sep, 1)
    if rest:
        _unflatten_dict(rest[0], value, out.setdefault(key, {}), sep)
    else:
        v, e, f = CifSerializer.string_to_variable(value)
        out[key] = v
        if e is not None:
            out['error'] = e
        if f is not None and not f:
            out['fixed'] = f

def _unflatten_loop(loop: cif.Loop, out: dict, sep: str):
    loop_name = "".join(p for p, *r in zip(*loop.tags) if all(p==c for c in r))
    labels = [lab[len(loop_name):] for lab in loop.tags]
    out[loop_name] = []
    width = loop.width()
    length = loop.length()
    for idx in range(length):
        this_dict = {}
        for idx2 in range(width):
            _unflatten_dict(labels[idx2], loop.val(idx, idx2), this_dict, sep)
        out[loop_name].append(this_dict)
    return out

def unflatten_bock(block: cif.Block, sep: str) -> dict:
    # Note that all names start with "_", so start at index 1!
    inter_dict = {item.pair[0][1:]: item.pair[1] for item in block if item.loop is None}
    loops = [loop.loop for loop in block if loop.loop is not None]
    out_dict = {}
    for k, v in inter_dict.items():
        _unflatten_dict(k, v, out_dict, sep)
    for loop in loops:
        _unflatten_loop(loop, out_dict, sep)
    return out_dict


class CifSerializer(BaseEncoderDecoder):

    def encode(self, obj: BV, skip: Optional[List[str]] = None, data_only: bool = False) -> str:
        if skip is None:
            skip = []
        doc = cif.Document()
        encoder = DictSerializer
        if data_only:
            encoder = DataDictSerializer
        obj_dict = obj.encode(encoder=encoder, skip=skip, include_id=False)
        flattened_obj = flatten_dict(obj_dict, sep=_SEP)
        block = doc.add_new_block(str(obj._borg.map.convert_id_to_key(obj)))
        for k, v in flattened_obj.items():
            self._check_class(doc, block, k, v)
        return doc.as_string()

    @classmethod
    def decode(cls, data_str) -> Union[BV, Dict[str, BV]]:
        document = cif.read_string(data_str)
        out_dict = {}
        for i, block in enumerate(document):
            out_dict[block.name] = DictSerializer.decode(unflatten_bock(block, sep=_SEP))
        if len(document) == 1:
            out_dict = list(out_dict.values())[0]
        return out_dict

    def _check_class(self, document: cif.Document, block: cif.Block, key: str, value: Any):
        T_ = type(value)
        if issubclass(T_, bool):
            new_obj = '@bool.' + str(value)
            self._str_encoder(block, key, new_obj)
        elif issubclass(T_, list):
            if len(value) > 0:
                test_item = value[0]
                if not issubclass(type(test_item), dict):
                    raise NotImplementedError
                else:
                    names = list(test_item.keys())
                    self._list_encoder(block, key, value, names)
        elif issubclass(T_, (str, Number)):
            self._str_encoder(block, key, value)

    def _str_encoder(self, block: cif.Block, key: str, in_str: str):
        in_str = self.variable_to_string(in_str)
        if key[0] != '_':
            key = '_' + key
        block.set_pair(key, in_str)

    def _list_encoder(self, block: cif.Block, name: str, items: List, keys: List[str]):
        if name[0] != '_':
            name = '_' + name
        loop: cif.Loop = block.init_loop(name, keys)
        for item in items:
            if len(keys) == 1:
                loop.add_row([self.variable_to_string(item)])
            else:
                loop.add_row([self.variable_to_string(item.get(key)) for key in keys])

    def variable_to_string(self, value, fixed: Optional[bool] = None, error: Optional[float] = None) -> str:
        decimal_places = 8
        if isinstance(value, Number):
            try:
                if np.isinf(value):
                    return self.variable_to_string(str(value))
            except TypeError:
                print(type(value))
            if abs(value - int(value)) > 0.0:
                decimal_places = len(str(value).split('.')[1]) - 1
            initial_str = "{:." + str(decimal_places) + "f}"
            s = initial_str.format(round(value, decimal_places))
            if error is not None and not np.isclose(error, 0.0):
                xe_exp = int(floor(log10(error)))

                # uncertainty
                un_exp = xe_exp - decimal_places + 1
                un_int = round(error * 10 ** (-un_exp))

                # nominal value
                no_exp = un_exp
                no_int = round(value * 10 ** (-no_exp))

                # format - nom(unc)
                fmt = "%%.%df" % max(0, -no_exp)
                s = (fmt + "(%.0f)") % (
                    no_int * 10 ** no_exp,
                    un_int * 10 ** max(0, un_exp),
                )
        elif isinstance(value, str):
            s = "{:s}".format(value)
        else:
            s = "{:s}".format(str(value))
        if fixed is not None and not fixed and error is None:
            s += "()"
        return self._format_field(s)

    @staticmethod
    def string_to_variable(in_string: str):
        in_string = in_string.strip()
        if "'" in in_string:
            in_string = in_string.replace("'", '')
        if "\"" in in_string:
            in_string = in_string.replace("\"", '')
        fixed = None
        error = None
        tokens = in_string.split("(")
        try:
            value = float(tokens[0])
        except ValueError:
            value = tokens[0]
            if value[0] == '@':
                tokens = value[1:].split('.')
                cls = tokens[0]
                val = tokens[1]
                value = eval(cls)(val)
            elif value == 'None':
                value = None
            return value, error, fixed
        if len(tokens) > 1:
            fixed = False
            if tokens[1][0] != ")":
                error = (10 ** -(len(f"{tokens[0]}".split(".")[1]) - 1)) * int(tokens[1][:-1])
        return value, error, fixed

    @staticmethod
    def _format_field(v):
        v = v.__str__().strip()
        if len(v) > _MAX_LEN:
            return ";\n" + textwrap.fill(v, _MAX_LEN) + "\n;"
        # add quotes if necessary
        if v == "":
            return '""'
        if (
            (" " in v or v[0] == "_")
            and not (v[0] == "'" and v[-1] == "'")
            and not (v[0] == '"' and v[-1] == '"')
        ):
            if "'" in v:
                q = '"'
            else:
                q = "'"
            v = q + v + q
        return v


from __future__ import annotations

__author__ = 'github.com/wardsimon'
__version__ = '0.0.1'

#  SPDX-FileCopyrightText: 2023 easyCrystallography contributors <crystallography@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2022-2023  Contributors to the easyCore project <https://github.com/easyScience/easyCrystallography>

from typing import List, NoReturn, TYPE_CHECKING, ClassVar, Tuple, Dict

from .template import CIF_Template, gemmi
from easyCrystallography.Components.Site import Site as _Site, Atoms as _Atoms
from easyCrystallography.Components.AtomicDisplacement import AtomicDisplacement as _AtomicDisplacement
from easyCrystallography.Components.Susceptibility import MagneticSusceptibility as _MagneticSusceptibility

if TYPE_CHECKING:
    from easyCore.Utils.typing import B


class AtomicDisplacement(CIF_Template):

    _CIF_SECTION_NAME: ClassVar[str] = "_atom_site"

    _CIF_ADP_ISO_CONVERSIONS = [
        ("label", "_label"),
        ("adp_type", "_adp_type"),
        ("Biso", "_B_iso_or_equiv"),
        ("Uiso", "_U_iso_or_equiv"),
    ]
    _CIF_ADP_ANISO_CONVERSIONS = [
        ("label", "_aniso_label"),
        ("B_11", "_B_11"),
        ("B_12", "_B_12"),
        ("B_13", "_B_13"),
        ("B_22", "_B_22"),
        ("B_23", "_B_23"),
        ("B_33", "_B_33"),
        ("U_11", "_U_11"),
        ("U_12", "_U_12"),
        ("U_13", "_U_13"),
        ("U_22", "_U_22"),
        ("U_23", "_U_23"),
        ("U_33", "_U_33"),
    ]

    def __init__(self, reference_class=_AtomicDisplacement):
        super().__init__()
        self._CIF_CLASS = reference_class

    def from_cif_block(self, block: gemmi.cif.Block) -> Dict[str, B]:

        atom_dict = {}
        # ADP CHECKER
        keys = [
            self._CIF_SECTION_NAME + name[1] if 'label' in name[1] else '?' + self._CIF_SECTION_NAME + name[1]
            for name in self._CIF_ADP_ISO_CONVERSIONS]
        table = block.find(keys)
        for row in table:
            kwargs = {}
            errors = {}
            is_fixed = {}
            if not row.has(0):
                continue
            if row.has(1):
                kwargs['adp_type'] = row[1]
            else:
                if row.has(2):
                    kwargs['adp_type'] = 'Biso'
                elif row.has(3):
                    kwargs['adp_type'] = 'Uiso'
                else:
                    continue

            for i in range(2, 4):
                if row.has(i):
                    V, E, F = self.string_to_variable(row[i])
                    kwargs[kwargs['adp_type']] = V
                    if E:
                        errors[kwargs['adp_type']] = E
                    if F is not None and not F:
                        is_fixed[kwargs['adp_type']] = F
            obj = _AtomicDisplacement(**kwargs)
            for error in errors.keys():
                setattr(getattr(obj, error), 'error', errors[error])
            for atr in is_fixed.keys():
                setattr(getattr(obj, atr), 'fixed', is_fixed[atr])
            atom_dict[row[0]] = {'adp': obj}

        keys = [
            self._CIF_SECTION_NAME + name[1] if 'label' in name[1] else '?' + self._CIF_SECTION_NAME + name[1]
            for name in self._CIF_ADP_ANISO_CONVERSIONS]
        table = block.find(keys)
        for row in table:
            kwargs = {}
            errors = {}
            is_fixed = {}
            if not row.has(0):
                continue
            if row.has(1):
                kwargs['adp_type'] = 'Bani'
                idx = 1
            elif row.has(7):
                idx = 7
                kwargs['adp_type'] = 'Uani'
            else:
                continue
            for i in range(idx, idx + 6):
                l = self._CIF_ADP_ANISO_CONVERSIONS[i][0]
                if row.has(i):
                    V, E, F = self.string_to_variable(row[i])
                    kwargs[l] = V
                    if E:
                        errors[l] = E
                    if F is not None and not F:
                        is_fixed[l] = F
            obj = _AtomicDisplacement(**kwargs)
            for error in errors.keys():
                setattr(getattr(obj, error), 'error', errors[error])
            for atr in is_fixed.keys():
                setattr(getattr(obj, atr), 'fixed', is_fixed[atr])
            atom_dict[row[0]] = {'adp': obj}
        return atom_dict

    def add_to_cif_block(self, obj: B, block: gemmi.cif.Block) -> NoReturn:
        labels = []
        objs = []

        # Assume that all atoms have the same adp type
        if len(obj) == 0:
            return labels, objs
        adp0 = getattr(obj[0], 'adp', None)
        if adp0 is None:
            return labels, objs
        adp_type = adp0.adp_type.raw_value
        if adp_type[-3:] == 'iso':
            labels = self._CIF_ADP_ISO_CONVERSIONS.copy()
            if adp_type[0].lower() == 'u':
                del labels[2]
            else:
                del labels[3]
            objs = [[getattr(_obj, 'adp')] for _obj in obj]
            del labels[0]
            return [[labels]]*len(objs), objs
        elif adp_type[-3:] == 'ani':
            labels = self._CIF_ADP_ANISO_CONVERSIONS.copy()
            if adp_type[0].lower() == 'u':
                idx = 7
            else:
                idx = 1
            rows = []
            for _obj in obj:
                row = [self.variable_to_string(getattr(_obj, self._CIF_ADP_ANISO_CONVERSIONS[0][0]))]
                for i in range(idx, idx + 6):
                    row.append(self.variable_to_string(getattr(_obj, self._CIF_ADP_ANISO_CONVERSIONS[i][0])))
                rows.append(row)
            del labels[idx: idx + 6]
            loop = block.init_loop(self._CIF_SECTION_NAME, [label[1] for label in labels])
            for row in rows:
                loop.add_row(row)
        return labels, objs

    def from_cif_string(self, cif_string: str) -> List[Dict[str, B]]:
        if "data_" not in cif_string:
            cif_string = "data_temp\n" + cif_string
        cif_blocks = gemmi.cif.read_string(cif_string)
        objs = []
        for block in cif_blocks:
            objs.append(self.from_cif_block(block))
        return objs


class MagneticSusceptibility(CIF_Template):

    _CIF_SECTION_NAME: ClassVar[str] = "_atom_site"
    _CIF_MSP_ANISO_CONVERSIONS = [
        ("label", "_susceptibility_label"),
        ("msp_type", "_susceptibility_chi_type"),
        ("chi_11", "_susceptibility_chi_11"),
        ("chi_12", "_susceptibility_chi_12"),
        ("chi_13", "_susceptibility_chi_13"),
        ("chi_22", "_susceptibility_chi_22"),
        ("chi_23", "_susceptibility_chi_23"),
        ("chi_33", "_susceptibility_chi_33"),
    ]

    def __init__(self, reference_class=_MagneticSusceptibility):
        super().__init__()
        self._CIF_CLASS = reference_class

    def from_cif_block(self, block: gemmi.cif.Block) -> Dict[str, B]:

        atom_dict = {}
        keys = [
            self._CIF_SECTION_NAME + name[1] if 'label' in name[1] else '?' + self._CIF_SECTION_NAME + name[1]
            for name in self._CIF_MSP_ANISO_CONVERSIONS]
        table = block.find(keys)
        for row in table:
            kwargs = {}
            errors = {}
            is_fixed = {}
            if not row.has(0):
                continue
            if row.has(1):
                kwargs['msp_type'] = row[1]
            else:
                continue
            idx = 2
            for i in range(idx, idx + 6):
                l = self._CIF_MSP_ANISO_CONVERSIONS[i][0]
                if row.has(i):
                    V, E, F = self.string_to_variable(row[i])
                    kwargs[l] = V
                    if E:
                        errors[l] = E
                    if F is not None and not F:
                        is_fixed[l] = F
            obj = _MagneticSusceptibility(**kwargs)
            for error in errors.keys():
                setattr(getattr(obj, error), 'error', errors[error])
            for atr in is_fixed.keys():
                setattr(getattr(obj, atr), 'fixed', is_fixed[atr])
            atom_dict[row[0]] = {'msp': obj}
        return atom_dict

    def add_to_cif_block(self, obj: B, block: gemmi.cif.Block) -> NoReturn:
        # Then add the additional loops
        lines = []
        for atom in obj:
            if not hasattr(atom, 'msp'):
                continue
            line = [self.variable_to_string(atom.__getattribute__(self._CIF_MSP_ANISO_CONVERSIONS[0][0]))]
            for keys in self._CIF_MSP_ANISO_CONVERSIONS[1:]:
                key, cif_key = keys
                s = self.variable_to_string(atom.msp.__getattribute__(key))
                line.append(s)
            lines.append(line)
        _, names = zip(*self._CIF_MSP_ANISO_CONVERSIONS)
        loop = block.init_loop(self._CIF_SECTION_NAME, names)
        for line in lines:
            loop.add_row(line)

    def from_cif_string(self, cif_string: str) -> List[B]:
        if "data_" not in cif_string:
            cif_string = "data_temp\n" + cif_string
        cif_blocks = gemmi.cif.read_string(cif_string)
        objs = []
        for block in cif_blocks:
            objs.append(self.from_cif_block(block))
        return objs


class Atoms(CIF_Template):

    _CIF_SECTION_NAME: ClassVar[str] = "_atom_site"
    _CIF_CONVERSIONS: ClassVar[List[Tuple[str, str]]] = [
        ("label", "_label"),
        ("specie", "_type_symbol"),
        ("fract_x", "_fract_x"),
        ("fract_y", "_fract_y"),
        ("fract_z", "_fract_z"),
        ("occupancy", "_occupancy"),
    ]

    def __init__(self, reference_class=_Atoms):
        super().__init__()
        self._CIF_CLASS = reference_class

    def _site_runner(self, block):
        keys = [
            self._CIF_SECTION_NAME + name[1] if 'occupancy' not in name[1] else '?' + self._CIF_SECTION_NAME + name[1]
            for name in self._CIF_CONVERSIONS]
        table = block.find(keys)
        atom_dict = {}
        error_dict = {}
        fixed_dict = {}
        for row in table:
            kwargs = {}
            errors = {}
            is_fixed = {}
            for idx, item in enumerate(self._CIF_CONVERSIONS):
                ec_name, cif_name = item
                if row.has(idx):
                    V, E, F = self.string_to_variable(row[idx])
                    kwargs[ec_name] = V
                    if E:
                        errors[ec_name] = E
                    if F is not None and not F:
                        is_fixed[ec_name] = F
            atom_dict[kwargs['label']] = kwargs
            error_dict[kwargs['label']] = errors
            fixed_dict[kwargs['label']] = is_fixed

        # ADP CHECKER
        ADP_RUNNER = AtomicDisplacement()
        adp_dict = ADP_RUNNER.from_cif_block(block)
        for label, adp in adp_dict.items():
            if label in atom_dict:
                atom_dict[label].update(adp)

        # MSP Checker
        MSP_RUNNER = MagneticSusceptibility()
        msp_dict = MSP_RUNNER.from_cif_block(block)
        for label, msp in msp_dict.items():
            if label in atom_dict:
                atom_dict[label].update(msp)

        atoms = []
        for a in atom_dict.values():
            obj = self._CIF_CLASS._SITE_CLASS(**a)
            if obj.label in error_dict.keys():
                for atr in error_dict[obj.label].keys():
                    setattr(getattr(obj, atr), 'error', error_dict[obj.label][atr])
            if obj.label in fixed_dict.keys():
                for atr in fixed_dict[obj.label].keys():
                    setattr(getattr(obj, atr), 'fixed', fixed_dict[obj.label][atr])
            atoms.append(obj)
        return atoms

    def from_cif_block(self, block: gemmi.cif.Block) -> B:
        atoms = self._site_runner(block)
        return self._CIF_CLASS('from_cif', *atoms)

    def add_to_cif_block(self, obj: B, block: gemmi.cif.Block) -> NoReturn:
        additional_keys = []
        additional_objs = []
        if len(obj) > 0:
            if getattr(obj[0], 'adp', False):
                ADP_WRITER = AtomicDisplacement()
                additional_keys, additional_objs = ADP_WRITER.add_to_cif_block(obj, block)
        MSP_WRITER = MagneticSusceptibility()
        MSP_WRITER.add_to_cif_block(obj, block)
        self._add_to_cif_block(obj, block, additional_keys, additional_objs)

    def _add_to_cif_block(self, obj: B, block: gemmi.cif.Block, additional_keys, additional_objs) -> NoReturn:
        # First add the main loop
        items = list(self._CIF_CONVERSIONS)
        names = [item[1] for item in items]
        lines = []
        for idx1, atom in enumerate(obj):
            line = []
            for idx2, item in enumerate(items):
                if item[0] == 'specie':
                    s = str(atom.specie)
                else:
                    s = self.variable_to_string(atom.__getattribute__(item[0]))
                line.append(s)
            lines.append(line)

        for keys, objs, line in zip(additional_keys, additional_objs, lines):
            for idx, _obj in enumerate(objs):
                for key in keys[idx]:
                    s = self.variable_to_string(_obj.__getattribute__(key[0]))
                    line.append(s)
                    if key[1] not in names:
                        names.append(key[1])
        loop = block.init_loop(self._CIF_SECTION_NAME, names)
        for line in lines:
            loop.add_row(line)

    def from_cif_string(self, cif_string: str) -> List[B]:
        if "data_" not in cif_string:
            cif_string = "data_temp\n" + cif_string
        cif_blocks = gemmi.cif.read_string(cif_string)
        objs = []
        for block in cif_blocks:
            objs.append(self.from_cif_block(block))
        return objs





#  SPDX-FileCopyrightText: 2023 easyCrystallography contributors <crystallography@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2022-2023  Contributors to the easyCore project <https://github.com/easyScience/easyCrystallography>


from __future__ import annotations

__author__ = 'github.com/wardsimon'
__version__ = '0.1.0'

from typing import List, Tuple, Union, ClassVar, TypeVar, Optional, TYPE_CHECKING

from easyCore import np
from easyCore.Utils.io.star import StarEntry, StarSection, StarLoop
from easyCore.Objects.ObjectClasses import BaseObj, Descriptor, Parameter
from easyCore.Utils.classTools import addProp, removeProp
from abc import abstractmethod

if TYPE_CHECKING:
    from easyCore.Utils.typing import iF

_ANIO_DETAILS = {
    'adp_type': {
        'description': "A standard code used to describe the type of atomic displacement parameters used for the site.",
        'url':         'https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Iatom_site_adp_type.html',
        'value':       'Uani'
    },
    'Uani':     {
        'description': 'Isotropic atomic displacement parameter, or equivalent isotropic atomic  displacement '
                       'parameter, U(equiv), in angstroms squared, calculated from anisotropic atomic displacement  '
                       'parameters.',
        'url':         'https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Iatom_site_aniso_U_.html',
        'value':       0.0,
        'units':       'angstrom^2',
        'fixed':       True,
    },
    'Uiso':     {
        'description': 'The standard anisotropic atomic displacement components in angstroms squared which appear in '
                       'the structure-factor term.',
        'url':         'https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Iatom_site_U_iso_or_equiv.html',
        'value':       0.0,
        'min':         0,
        'max':         np.inf,
        'units':       'angstrom^2',
        'fixed':       True,
    },
    'Bani':     {
        'description': 'The standard anisotropic atomic displacement components in angstroms squared which appear in '
                       'the structure-factor term.',
        'url':         'https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Iatom_site_aniso_B_.html',
        'value':       0.0,
        'units':       'angstrom^2',
        'fixed':       True,
    },
    'Biso':     {
        'description': 'Isotropic atomic displacement parameter, or equivalent isotropic atomic displacement '
                       'parameter, B(equiv), in angstroms squared, calculated from anisotropic displacement '
                       'components.',
        'url':         'https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Iatom_site_B_iso_or_equiv.html',
        'value':       0.0,
        'min':         0,
        'max':         np.inf,
        'units':       'angstrom^2',
        'fixed':       True,
    }
}


class AdpBase(BaseObj):

    def __init__(self, *args, **kwargs):
        super(AdpBase, self).__init__(*args, **kwargs)

    @property
    def matrix(self) -> np.ndarray:
        matrix = np.zeros([3, 3])
        pars = self.get_parameters()
        if len(pars) == 1:
            np.fill_diagonal(matrix, pars[0].raw_value)
        elif len(pars) == 6:
            matrix[0, 0] = pars[0].raw_value
            matrix[0, 1] = pars[1].raw_value
            matrix[0, 2] = pars[2].raw_value
            matrix[1, 1] = pars[3].raw_value
            matrix[1, 2] = pars[4].raw_value
            matrix[2, 2] = pars[5].raw_value
        return matrix

    def __repr__(self):
        s = f'{self.name} - ('
        for par in self.get_parameters():
            s += f'{par.name}: {par.raw_value}, '
        s = s[:-2] + ')'
        return s


class Anisotropic(AdpBase):

    U_11: ClassVar[Parameter]
    U_12: ClassVar[Parameter]
    U_13: ClassVar[Parameter]
    U_22: ClassVar[Parameter]
    U_23: ClassVar[Parameter]
    U_33: ClassVar[Parameter]

    def __init__(self,
                 U_11: Optional[Union[Parameter, float]] = None, U_12: Optional[Union[Parameter, float]] = None,
                 U_13: Optional[Union[Parameter, float]] = None, U_22: Optional[Union[Parameter, float]] = None,
                 U_23: Optional[Union[Parameter, float]] = None, U_33: Optional[Union[Parameter, float]] = None,
                 interface: Optional[iF] = None):
        super(Anisotropic, self).__init__('anisoU',
                                          U_11=Parameter('U_11', **_ANIO_DETAILS['Uani']), U_12=Parameter('U_12', **_ANIO_DETAILS['Uani']),
                                          U_13=Parameter('U_13', **_ANIO_DETAILS['Uani']), U_22=Parameter('U_22', **_ANIO_DETAILS['Uani']),
                                          U_23=Parameter('U_23', **_ANIO_DETAILS['Uani']), U_33=Parameter('U_33', **_ANIO_DETAILS['Uani']))
        if U_11 is not None:
            self.U_11 = U_11
        if U_12 is not None:
            self.U_12 = U_12
        if U_13 is not None:
            self.U_13 = U_13
        if U_22 is not None:
            self.U_22 = U_22
        if U_23 is not None:
            self.U_23 = U_23
        if U_33 is not None:
            self.U_33 = U_33
        self.interface = interface


class Isotropic(AdpBase):

    Uiso: ClassVar[Parameter]

    def __init__(self, Uiso: Optional[Union[Parameter, float]] = None, interface: Optional[iF] = None):
        super(Isotropic, self).__init__('Uiso',
                                        Uiso=Parameter('Uiso', **_ANIO_DETAILS['Uiso']))
        if Uiso is not None:
            self.Uiso = Uiso
        self.interface = interface


class AnisotropicBij(AdpBase):

    B_11: ClassVar[Parameter]
    B_12: ClassVar[Parameter]
    B_13: ClassVar[Parameter]
    B_22: ClassVar[Parameter]
    B_23: ClassVar[Parameter]
    B_33: ClassVar[Parameter]

    def __init__(self,
                 B_11: Optional[Union[Parameter, float]] = None, B_12: Optional[Union[Parameter, float]] = None,
                 B_13: Optional[Union[Parameter, float]] = None, B_22: Optional[Union[Parameter, float]] = None,
                 B_23: Optional[Union[Parameter, float]] = None, B_33: Optional[Union[Parameter, float]] = None,
                 interface: Optional[iF] = None):
        super(AnisotropicBij, self).__init__('anisoB',
                                             **{name: Parameter(name, **_ANIO_DETAILS['Bani']) for name in
                                              ['B_11', 'B_12', 'B_13',
                                               'B_22', 'B_23', 'B_33']
                                                })
        if B_11 is not None:
            self.B_11 = B_11
        if B_12 is not None:
            self.B_12 = B_12
        if B_13 is not None:
            self.B_13 = B_13
        if B_22 is not None:
            self.B_22 = B_22
        if B_23 is not None:
            self.B_23 = B_23
        if B_33 is not None:
            self.B_33 = B_33
        self.interface = interface


class IsotropicB(AdpBase):

    Biso: ClassVar[Parameter]

    def __init__(self, Biso: Optional[Union[Parameter, float]] = None, interface: Optional[iF] = None):
        super(IsotropicB, self).__init__('Biso',
                                         Biso=Parameter('Biso', **_ANIO_DETAILS['Biso']))
        if Biso is not None:
            self.Biso = Biso
        self.interface = interface


_AVAILABLE_ISO_TYPES = {
    'Uani': Anisotropic,
    'Uiso': Isotropic,
    # 'Uovl': 'Overall',
    # 'Umpe': 'MultipoleExpansion',
    'Bani': AnisotropicBij,
    'Biso': IsotropicB,
    # 'Bovl': 'OverallB'
}

if TYPE_CHECKING:
    AB = TypeVar('AB', bound=AdpBase)


class AtomicDisplacement(BaseObj):

    adp_type: ClassVar[Descriptor]
    adp_class: ClassVar[AB]

    def __init__(self, adp_type: Optional[Union[Descriptor, str]] = None, interface: Optional[iF] = None, **kwargs):
        if adp_type is None:
            adp_type = 'Uiso'
        if isinstance(adp_type, str):
            adp_type = Descriptor('adp_type', adp_type)
        adp_class_name = adp_type.raw_value
        if adp_class_name in _AVAILABLE_ISO_TYPES.keys():
            adp_class = _AVAILABLE_ISO_TYPES[adp_class_name]
            # enable passing ADP parameters to constructor
            if "adp_class" in kwargs.keys():
                m = getattr(kwargs["adp_class"], adp_class_name)
                kwargs[adp_class_name] = m
                _ = kwargs.pop("adp_class")
            adp = adp_class(**kwargs, interface=interface)
        else:
            raise AttributeError(f"{adp_class_name} is not a valid adp type")
        super(AtomicDisplacement, self).__init__('adp',
                                                 adp_type=adp_type,
                                                 adp_class=adp)
        for par in adp.get_parameters():
            addProp(self, par.name, fget=self.__a_getter(par.name), fset=self.__a_setter(par.name))
        self.interface = interface

    def switch_type(self, adp_string: str, **kwargs):
        if adp_string in _AVAILABLE_ISO_TYPES.keys():
            adp_class = _AVAILABLE_ISO_TYPES[adp_string]
            if kwargs:
                adp = adp_class(**kwargs, interface=self.interface)
            else:
                adp = adp_class(interface=self.interface)
        else:
            raise AttributeError(f"{adp_string} is not a valid adp type")
        for par in self.adp_class.get_parameters():
            removeProp(self, par.name)
        self.adp_class = adp_class
        self.adp_type = adp
        for par in adp.get_parameters():
            addProp(self, par.name, fget=self.__a_getter(par.name), fset=self.__a_setter(par.name))

    @property
    def available_types(self) -> List[str]:
        return [name for name in _AVAILABLE_ISO_TYPES.keys()]

    @staticmethod
    def __a_getter(key: str):

        def getter(obj):
            return obj.adp_class._kwargs[key]

        return getter

    @staticmethod
    def __a_setter(key):
        def setter(obj, value):
            obj.adp_class._kwargs[key].value = value

        return setter



#  SPDX-FileCopyrightText: 2023 easyCrystallography contributors <crystallography@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2022-2023  Contributors to the easyCore project <https://github.com/easyScience/easyCrystallography>


__author__ = 'github.com/wardsimon'
__version__ = '0.1.0'



#  SPDX-FileCopyrightText: 2023 easyCrystallography contributors <crystallography@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2022-2023  Contributors to the easyCore project <https://github.com/easyScience/easyCrystallography>

from __future__ import annotations

__author__ = 'github.com/wardsimon'
__version__ = '0.1.0'

import collections
import itertools
import math
import warnings

from fractions import Fraction
from functools import reduce
from typing import (
    Tuple,
    Union,
    List,
    Sequence,
    Dict,
    ClassVar,
    TypeVar,
    Optional,
    TYPE_CHECKING,
    Type,
)

from easyCore import np, ureg
from easyCore.Fitting.Constraints import ObjConstraint
from easyCore.Objects.ObjectClasses import Parameter, BaseObj
from easyCore.Utils.decorators import memoized
from .SpaceGroup import SpaceGroup
from copy import deepcopy

Vector3Like = Union[List[float], np.ndarray]


if TYPE_CHECKING:
    from easyCore.Utils.typing import iF

CELL_DETAILS = {
    "length": {
        "description": "Unit-cell length of the selected structure in angstroms.",
        "url": "https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Icell_length_.html",
        "value": 3,
        "units": "angstrom",
        "min": 0,
        "max": np.Inf,
        "fixed": True,
    },
    "angle": {
        "description": "Unit-cell angle of the selected structure in degrees.",
        "url": "https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Icell_angle_.html",
        "value": 90,
        "units": "deg",
        "min": 0,
        "max": np.Inf,
        "fixed": True,
    },
}


class Lattice(BaseObj):

    _REDIRECT = {
        'ang_unit': None
    }

    length_a: ClassVar[Parameter]
    length_b: ClassVar[Parameter]
    length_c: ClassVar[Parameter]
    angle_alpha: ClassVar[Parameter]
    angle_beta: ClassVar[Parameter]
    angle_gamma: ClassVar[Parameter]

    def __init__(
        self,
        length_a: Optional[Union[Parameter, float]] = None,
        length_b: Optional[Union[Parameter, float]] = None,
        length_c: Optional[Union[Parameter, float]] = None,
        angle_alpha: Optional[Union[Parameter, float]] = None,
        angle_beta: Optional[Union[Parameter, float]] = None,
        angle_gamma: Optional[Union[Parameter, float]] = None,
        interface: Optional[iF] = None,
        ang_unit: str = 'deg',
        **kwargs,
    ):

        THESE_CELL_DETAILS = deepcopy(CELL_DETAILS)
        THESE_CELL_DETAILS['angle']['units'] = ang_unit

        super().__init__(
            "lattice",
            length_a=Parameter("length_a", **THESE_CELL_DETAILS["length"]),
            length_b=Parameter("length_b", **THESE_CELL_DETAILS["length"]),
            length_c=Parameter("length_c", **THESE_CELL_DETAILS["length"]),
            angle_alpha=Parameter("angle_alpha", **THESE_CELL_DETAILS["angle"]),
            angle_beta=Parameter("angle_beta", **THESE_CELL_DETAILS["angle"]),
            angle_gamma=Parameter("angle_gamma", **THESE_CELL_DETAILS["angle"]),
            **kwargs
        )
        if length_a is not None:
            self.length_a = length_a
        if length_b is not None:
            self.length_b = length_b
        if length_c is not None:
            self.length_c = length_c

        if angle_alpha is not None:
            self.angle_alpha = angle_alpha
        if angle_beta is not None:
            self.angle_beta = angle_beta
        if angle_gamma is not None:
            self.angle_gamma = angle_gamma

        if THESE_CELL_DETAILS['angle']['units'] is not CELL_DETAILS['angle']['units']:
            self.angle_alpha.convert_unit(CELL_DETAILS['angle']['units'])
            self.angle_beta.convert_unit(CELL_DETAILS['angle']['units'])
            self.angle_gamma.convert_unit(CELL_DETAILS['angle']['units'])

        self.interface = interface

    @classmethod
    def from_matrix(
        cls,
        matrix: Union[List[float], List[List[float]], np.ndarray],
        interface: Optional[iF] = None,
    ) -> L:
        """
        Construct a crystallographic unit cell from the lattice matrix

        :param matrix: Matrix describing the crystallographic unit cell in the form of a numpy array,
        1x9 list or 3x3 list.
        :type matrix: np.ndarray, List[float], List[List[float]]
        :return: Crystallographic unit cell container
        :rtype: Lattice
        """

        matrix = np.array(matrix, dtype=np.float64).reshape((3, 3))
        lengths = np.sqrt(np.sum(matrix ** 2, axis=1))
        angles = np.zeros(3)
        for i in range(3):
            j = (i + 1) % 3
            k = (i + 2) % 3
            angles[i] = np.dot(matrix[j], matrix[k]) / (lengths[j] * lengths[k])
            angles[i] = max(min(angles[i], 1), -1)
        angles = np.arccos(angles) * 180.0 / np.pi
        return cls(*lengths, *angles, interface=interface)

    @classmethod
    def cubic(cls, a: float, interface: Optional[iF] = None) -> L:
        """
        Convenience constructor for a cubic lattice.

        :param a: The *a* lattice parameter of the cubic cell.
        :type a: float
        :return: Crystallographic unit cell container
        :rtype: Lattice
        """
        return cls(a, a, a, 90, 90, 90, interface=interface)

    @classmethod
    def tetragonal(cls, a: float, c: float, interface: Optional[iF] = None) -> L:
        """
        Convenience constructor for a tetragonal lattice.

        :param a: *a* lattice parameter of the tetragonal cell.
        :type a: float
        :param c: *c* lattice parameter of the tetragonal cell.
        :type c: float
        :return: Crystallographic unit cell container
        :rtype: Lattice
        """
        return cls(a, a, c, 90, 90, 90, interface=interface)

    @classmethod
    def orthorhombic(
        cls, a: float, b: float, c: float, interface: Optional[iF] = None
    ) -> L:
        """
        Convenience constructor for an orthorhombic lattice.

        :param a: *a* lattice parameter of the orthorhombic cell.
        :type a: float
        :param b: *b* lattice parameter of the orthorhombic cell.
        :type b: float
        :param c: *c* lattice parameter of the orthorhombic cell.
        :type c: float
        :return: Crystallographic unit cell container
        :rtype: Lattice
        """
        return cls(a, b, c, 90, 90, 90, interface=interface)

    @classmethod
    def monoclinic(
        cls,
        a: float,
        b: float,
        c: float,
        beta: float,
        interface: Optional[iF] = None,
    ) -> L:
        """
        Convenience constructor for a monoclinic lattice of dimensions a x b x c with non right-angle
        beta between lattice vectors a and c.

        :param a: *a* lattice parameter of the monoclinc cell.
        :type a: float
        :param b: *b* lattice parameter of the monoclinc cell.
        :type b: float
        :param c: *c* lattice parameter of the monoclinc cell.
        :type c: float
        :param beta: *beta* angle between lattice vectors b and c in degrees.
        :type beta: float
        :return: Crystallographic unit cell container
        :rtype: Lattice
        """
        return cls(a, b, c, 90, beta, 90, interface=interface)

    @classmethod
    def hexagonal(cls, a: float, c: float, interface: Optional[iF] = None) -> L:
        """
        Convenience constructor for a hexagonal lattice of dimensions a x a x c

        :param a: *a* lattice parameter of the hexagonal cell.
        :type a: float
        :param c: *c* lattice parameter of the hexagonal cell.
        :type c: float
        :return: Crystallographic unit cell container
        :rtype: Lattice
        """
        return cls(a, a, c, 90, 90, 120, interface=interface)

    @classmethod
    def rhombohedral(
        cls, a: float, alpha: float, interface: Optional[iF] = None
    ) -> L:
        """
        Convenience constructor for a rhombohedral lattice of dimensions a x a x a.

        :param a: *a* lattice parameter of the monoclinc cell.
        :type a: float
        :param alpha: Angle for the rhombohedral lattice in degrees.
        :type alpha: float
        :return: Crystallographic unit cell container
        :rtype: Lattice
        """
        return cls(a, a, a, alpha, alpha, alpha, interface=interface)

    # Dynamic properties
    @property
    def a(self) -> float:
        """
        Get the *a* lattice parameter.

        :return: *a* lattice parameter
        :rtype: float
        """
        return self.length_a.raw_value

    @a.setter
    def a(self, new_a_value: float):
        """
        Set the *a* lattice parameter.

        :param new_a_value: new *a* lattice parameter
        :type new_a_value: float
        :return: noneType
        :rtype: None
        """
        self.length_a.value = new_a_value

    @property
    def b(self) -> float:
        """
        Get the *b* lattice parameter.

        :return: *b* lattice parameter
        :rtype: float
        """
        return self.length_b.raw_value

    @b.setter
    def b(self, new_b_value: float):
        """
        Set the *b* lattice parameter.

        :param new_b_value: new *a* lattice parameter
        :type new_b_value: float
        :return: noneType
        :rtype: None
        """
        self.length_b.value = new_b_value

    @property
    def c(self) -> float:
        """
        Get the *c* lattice parameter.

        :return: *c* lattice parameter
        :rtype: float
        """
        return self.length_c.raw_value

    @c.setter
    def c(self, new_c_value: float):
        """
        Set the *c* lattice parameter.

        :param new_c_value: new *a* lattice parameter
        :type new_c_value: float
        :return: noneType
        :rtype: None
        """
        self.length_c.value = new_c_value

    @property
    def alpha(self) -> float:
        """
        Get the *alpha* lattice parameter.

        :return: *alpha* lattice parameter
        :rtype: float
        """
        return self.angle_alpha.raw_value

    @alpha.setter
    def alpha(self, new_alpha_value: float):
        """
        Set the *alpha* lattice parameter.

        :param new_alpha_value: new *alpha* lattice parameter
        :type new_alpha_value: float
        :return: noneType
        :rtype: None
        """
        self.angle_alpha.value = new_alpha_value

    @property
    def beta(self) -> float:
        """
        Get the *beta* lattice parameter.

        :return: *beta* lattice parameter
        :rtype: float
        """
        return self.angle_beta.raw_value

    @beta.setter
    def beta(self, new_beta_value: float):
        """
        Set the *beta* lattice parameter.

        :param new_beta_value: new *a* lattice parameter
        :type new_beta_value: float
        :return: noneType
        :rtype: None
        """
        self.angle_beta.value = new_beta_value

    @property
    def gamma(self) -> float:
        """
        Get the *gamma* lattice parameter.

        :return: *gamma* lattice parameter
        :rtype: float
        """
        return self.angle_gamma.raw_value

    @gamma.setter
    def gamma(self, new_gamma_value: float):
        """
        Set the *gamma* lattice parameter.

        :param new_gamma_value: new *gamma* lattice parameter
        :type new_gamma_value: float
        :return: noneType
        :rtype: None
        """
        self.angle_gamma.value = new_gamma_value

    @property
    def lengths(self) -> Tuple[float, float, float]:
        """
        Get the lengths of the unit cell.

        :return: Tuple of unit cell lengths (a, b, c)
        :rtype: tuple
        """
        return self.a, self.b, self.c

    @property
    def angles(self) -> Tuple[float, float, float]:
        """
        Get the angles of the unit cell in degrees.

        :return: Tuple of unit cell angles (alpha, beta, gamma)
        :rtype: tuple
        """
        return self.alpha, self.beta, self.gamma

    @property
    def matrix(self) -> np.ndarray:
        """
        Get the lattice matrix.

        :return: Lattice matrix in the form of a 9x9 matrix
        :rtype: np.ndarray
        """
        return self.__matrix(*self.lengths, *self.angles)

    @property
    def volume(self) -> float:
        """
        Volume of the unit cell.

        :return: Volume of the unit cell
        :rtype: ureg.Quantity
        """
        m = self.matrix
        vol = float(abs(np.dot(np.cross(m[0], m[1]), m[2])))
        return ureg.Quantity(
            vol, units=self.length_a.unit * self.length_b.unit * self.length_c.unit
        )

    @property
    def inv_matrix(self) -> np.ndarray:
        """
        Inverse of lattice matrix.

        :return: Inverse of lattice matrix
        :rtype: np.ndarray
        """
        return np.linalg.inv(self.matrix)

    @property
    def metric_tensor(self) -> np.ndarray:
        """
        The metric tensor of the lattice.

        :return metric tensor of the lattice
        :rtype: np.ndarray
        """
        matrix = self.matrix
        return np.dot(matrix, self.matrix.T)

    # Functions that create new copies
    @property
    def reciprocal_lattice(self) -> L:
        """
        Return the reciprocal lattice. Note that this is the standard
        reciprocal lattice used for solid state physics with a factor of 2 *
        pi. If you are looking for the crystallographic reciprocal lattice,
        use the reciprocal_lattice_crystallographic property.

        :return: New cell in the reciprocal lattice
        :rtype: Lattice
        """
        v = np.linalg.inv(self.matrix).T
        return self.__class__.from_matrix(v * 2 * np.pi, interface=self.interface)

    @property
    def reciprocal_lattice_crystallographic(self) -> L:
        """
        Returns the *crystallographic* reciprocal lattice, i.e., no factor of
        2 * pi.

        :return: New cell in the *crystallographic* reciprocal lattice
        :rtype: Lattice
        """
        return self.__class__.from_matrix(
            self.reciprocal_lattice.matrix / (2 * np.pi), interface=self.interface
        )

    def scale(self, new_volume: float) -> L:
        """
        Return a new Cell with volume new_volume by performing a
        scaling of the lattice vectors so that length proportions and angles
        are preserved.

        :param new_volume: New volume to scale to.
        :type new_volume: float
        :return: New cell scaled to volume
        :rtype: Lattice
        """

        lengths = self.lengths
        versors = self.matrix / lengths
        geo_factor = np.abs(np.dot(np.cross(versors[0], versors[1]), versors[2]))
        ratios = np.array(lengths) / lengths[2]
        new_c = (new_volume / (geo_factor * np.prod(ratios))) ** (1 / 3.0)
        return self.__class__.from_matrix(
            versors * (new_c * ratios), interface=self.interface
        )

    def scale_lengths(self, length_scales: Union[float, Vector3Like]) -> L:
        """
        Return a new Cell where the cell lengths have been scaled by the scaling
        factors. Angles are preserved.

        :param length_scales: Scaling for each length or universal scaling factor
        :type length_scales: Union[float, Vector3Like]
        :return: New Cell scaled by user supplied scale factors
        :rtype: Lattice
        """
        if isinstance(length_scales, float):
            length_scales = 3 * [length_scales]
        new_lengths = np.array(length_scales) * np.array(self.lengths)
        return self.__class__(
            *new_lengths, *self.angles, interface=self.interface
        )

    # Get functions
    def get_cartesian_coords(self, fractional_coords: Vector3Like) -> np.ndarray:
        """
        Returns the cartesian coordinates given fractional coordinates.

        :param fractional_coords: fractional_coords (3x1 array): Fractional coords.
        :type fractional_coords: np.ndarray , list
        :return: Cartesian coordinates
        :rtype: np.ndarray
        """
        return np.dot(fractional_coords, self.matrix)

    def get_fractional_coords(self, cart_coords: Vector3Like) -> np.ndarray:
        """
        Returns the fractional coordinates given cartesian coordinates.

        :param cart_coords: cart_coords (3x1 array): Fractional coords.
        :type cart_coords: np.ndarray , list
        :return: Fractional coordinates.
        :rtype: np.ndarray
        """
        return np.dot(cart_coords, self.inv_matrix)

    def get_vector_along_lattice_directions(
        self, cart_coords: Vector3Like
    ) -> np.ndarray:
        """
        Returns the coordinates along lattice directions given cartesian coordinates.
        Note, this is different than a projection of the cartesian vector along the
        lattice parameters. It is simply the fractional coordinates multiplied by the
        lattice vector magnitudes.

        :param cart_coords: cart_coords (3x1 array): Fractional coords.
        :type cart_coords: np.ndarray , list
        :return: Lattice coordinates.
        :rtype: np.ndarray
        """
        return self.lengths * self.get_fractional_coords(cart_coords)

    def d_hkl(self, miller_index: Vector3Like) -> float:
        """
        Returns the distance between the hkl plane and the origin

        :param miller_index: (h. k. l) Miller index of plane
        :type miller_index: Vector3Like
        :return: Distance between the hkl plane and the origin
        :rtype: float
        """

        gstar = self.reciprocal_lattice_crystallographic.metric_tensor
        hkl = np.array(miller_index)
        return 1 / ((np.dot(np.dot(hkl, gstar), hkl.T)) ** (1 / 2))

    # Checking
    def is_orthogonal(self) -> bool:
        """
        Returns true if the cell is orthogonal (all angles 90 degrees)

        :return: Whether all angles are 90 degrees.
        :rtype: bool
        """
        return all([abs(a - 90) < 1e-5 for a in self.angles])

    def is_hexagonal(
        self, hex_angle_tol: float = 5, hex_length_tol: float = 0.01
    ) -> bool:
        """
        Returns true if the Lattice is hexagonal.

        :param hex_angle_tol: Angle tolerance
        :param hex_length_tol: Length tolerance
        :return: Whether lattice corresponds to hexagonal lattice.
        :rtype: bool
        """
        lengths = self.lengths
        angles = self.angles
        right_angles = [i for i in range(3) if abs(angles[i] - 90) < hex_angle_tol]
        hex_angles = [
            i
            for i in range(3)
            if abs(angles[i] - 60) < hex_angle_tol
            or abs(angles[i] - 120) < hex_angle_tol
        ]

        return (
            len(right_angles) == 2
            and len(hex_angles) == 1
            and abs(lengths[right_angles[0]] - lengths[right_angles[1]])
            < hex_length_tol
        )

    @staticmethod
    @memoized
    def __matrix(
        a: float, b: float, c: float, alpha: float, beta: float, gamma: float
    ) -> np.ndarray:
        """
        Calculating the crystallographic matrix is time consuming and we use it often, so we have memoized it.
        :param a: *a* lattice parameter
        :type a: float
        :param b: *b* lattice parameter
        :type b: float
        :param c: *c* lattice parameter
        :type c: float
        :param alpha: *alpha* lattice parameter
        :type alpha: float
        :param beta: *beta* lattice parameter
        :type beta: float
        :param gamma: *gamma* lattice parameter
        :type gamma: float
        :return: crystallographic matrix
        :rtype: np.ndarray
        """
        angles_r = np.radians([alpha, beta, gamma])
        cos_alpha, cos_beta, cos_gamma = np.cos(angles_r)
        sin_alpha, sin_beta, sin_gamma = np.sin(angles_r)

        val = (cos_alpha * cos_beta - cos_gamma) / (sin_alpha * sin_beta)
        # Sometimes rounding errors result in values slightly > 1.
        val = max(min(val, 1), -1)
        gamma_star = np.arccos(val)

        vector_a = [a * sin_beta, 0.0, a * cos_beta]
        vector_b = [
            -b * sin_alpha * np.cos(gamma_star),
            b * sin_alpha * np.sin(gamma_star),
            b * cos_alpha,
        ]
        vector_c = [0.0, 0.0, float(c)]
        return np.array([vector_a, vector_b, vector_c], dtype=np.float64)

    # noinspection PyStringFormat
    def __repr__(self) -> str:
        return (
            "<Lattice: (a: {:.2f} {:~P}, b: {:.2f} {:~P}, c: {:.2f}{:~P}, alpha: {:.2f} {:~P}, beta: {:.2f} {:~P}, "
            ""
            "gamma: {:.2f} {:~P}>".format(
                self.length_a.raw_value,
                self.length_a.unit,
                self.length_b.raw_value,
                self.length_b.unit,
                self.length_c.raw_value,
                self.length_c.unit,
                self.angle_alpha.raw_value,
                self.angle_alpha.unit,
                self.angle_beta.raw_value,
                self.angle_beta.unit,
                self.angle_gamma.raw_value,
                self.angle_gamma.unit,
            )
        )

    def __format__(self, fmt_spec=""):
        """
        Support format printing. Supported formats are:

        1. "l" for a list format that can be easily copied and pasted, e.g.,
           ".3fl" prints something like
           "[[10.000, 0.000, 0.000], [0.000, 10.000, 0.000], [0.000, 0.000, 10.000]]"
        2. "p" for lattice parameters ".1fp" prints something like
           "{10.0, 10.0, 10.0, 90.0, 90.0, 90.0}"
        3. Default will simply print a 3x3 matrix form. E.g.,
           10.000 0.000 0.000
           0.000 10.000 0.000
           0.000 0.000 10.000
        """
        m = (self.lengths, self.angles)

        if fmt_spec.endswith("m"):
            fmt = "[[{}, {}, {}], [{}, {}, {}], [{}, {}, {}]]"
            m = self.matrix.tolist()
            fmt_spec = fmt_spec[:-1]
        elif fmt_spec.endswith("l"):
            fmt = "{{{}, {}, {}, {}, {}, {}}}"
            fmt_spec = fmt_spec[:-1]
        else:
            fmt = "({} {} {}), ({} {} {})"
            fmt_spec = fmt_spec[:-1]
        return fmt.format(*[format(c, fmt_spec) for row in m for c in row])

    def __copy__(self):
        """
        Returns a deep copy of the Lattice


        :return: Deep copy of self
        :rtype:
        """
        return self.__class__(
            *self.lengths, *self.angles, interface=self.interface
        )

    def get_points_in_sphere(
        self,
        frac_points: List[Vector3Like],
        center: Vector3Like,
        r: float,
        zip_results=True,
    ) -> Union[List[Tuple[np.ndarray, float, int, np.ndarray]], List[np.ndarray],]:
        """
        Find all points within a sphere from the point taking into account
        periodic boundary conditions. This includes sites in other periodic
        images.
        Algorithm:
        1. place sphere of radius r in crystal and determine minimum supercell
           (parallelpiped) which would contain a sphere of radius r. for this
           we need the projection of a_1 on a unit vector perpendicular
           to a_2 & a_3 (i.e. the unit vector in the direction b_1) to
           determine how many a_1"s it will take to contain the sphere.
           Nxmax = r * length_of_b_1 / (2 Pi)
        2. keep points falling within r.
        Args:
            frac_points: All points in the lattice in fractional coordinates.
            center: Cartesian coordinates of center of sphere.
            r: radius of sphere.
            zip_results (bool): Whether to zip the results together to group by
                 point, or return the raw fcoord, dist, index arrays
        Returns:
            if zip_results:
                [(fcoord, dist, index, supercell_image) ...] since most of the time, subsequent
                processing requires the distance, index number of the atom, or index of the image
            else:
                fcoords, dists, inds, image
        """
        cart_coords = self.get_cartesian_coords(frac_points)
        neighbors = get_points_in_spheres(
            all_coords=cart_coords,
            center_coords=np.array([center]),
            r=r,
            pbc=True,
            numerical_tol=1e-8,
            lattice=self,
            return_fcoords=True,
        )[0]
        if len(neighbors) < 1:
            return [] if zip_results else [()] * 4
        if zip_results:
            return neighbors
        return [np.array(i) for i in list(zip(*neighbors))]


L = TypeVar("L", bound=Lattice)


class PeriodicLattice(Lattice):

    spacegroup: ClassVar[SpaceGroup]

    def __init__(
        self,
        length_a: Optional[Union[Parameter, float]] = None,
        length_b: Optional[Union[Parameter, float]] = None,
        length_c: Optional[Union[Parameter, float]] = None,
        angle_alpha: Optional[Union[Parameter, float]] = None,
        angle_beta: Optional[Union[Parameter, float]] = None,
        angle_gamma: Optional[Union[Parameter, float]] = None,
        spacegroup: Optional[Union[SpaceGroup, str]] = None,
        interface: Optional[iF] = None
    ):
        super().__init__(
            length_a=length_a,
            length_b=length_b,
            length_c=length_c,
            angle_alpha=angle_alpha,
            angle_beta=angle_beta,
            angle_gamma=angle_gamma,
            spacegroup=SpaceGroup('P1'),
        )
        if spacegroup is not None:
            if isinstance(spacegroup, str):
                spacegroup = SpaceGroup(spacegroup)
            self.spacegroup = spacegroup

        # Do some class voodoo. We can do this as we have hidden space_group_HM_name
        # as _space_group_HM_name and used a simple property.
        self.__previous_SG_setter = spacegroup.__class__.space_group_HM_name.fset
        spacegroup.__class__.space_group_HM_name = property(
            fget=spacegroup.__class__.space_group_HM_name.fget,
            fset=lambda obj, val: self.__new_SG_setter(obj, val),
            fdel=spacegroup.__class__.space_group_HM_name.fdel,
        )

        self.interface = interface
        self.enforce_sym()

    @classmethod
    def from_lattice_and_spacegroup(
        cls: Type[L], lattice: Lattice, spacegroup: SpaceGroup
    ):
        length_a = lattice.length_a.raw_value
        length_b = lattice.length_b.raw_value
        length_c = lattice.length_c.raw_value
        angle_alpha = lattice.angle_alpha.raw_value
        angle_beta = lattice.angle_beta.raw_value
        angle_gamma = lattice.angle_gamma.raw_value

        return cls(
            length_a,
            length_b,
            length_c,
            angle_alpha,
            angle_beta,
            angle_gamma,
            spacegroup,
            interface=lattice.interface,
        )

    @classmethod
    def from_matrix(
        cls: Type[L],
        matrix: Union[List[float], List[List[float]], np.ndarray],
        interface: Optional[iF] = None
    ) -> L:
        """
        Construct a crystallographic unit cell from the lattice matrix

        :param matrix: Matrix describing the crystallographic unit cell in the form of a numpy array,
        1x9 list or 3x3 list.
        :type matrix: np.ndarray, List[float], List[List[float]]
        :return: Crystallographic unit cell container
        :rtype: Lattice
        """
        raise NotImplementedError(
            "A periodic Lattice can not be created from just a matrix"
        )

    def clear_sym(self):
        pars = [
            self.length_a,
            self.length_b,
            self.length_c,
            self.angle_alpha,
            self.angle_beta,
            self.angle_gamma,
        ]
        for par in pars:
            new_con = {
                con: par.user_constraints[con]
                for con in par.user_constraints.keys()
                if not con.startswith("sg_")
            }
            par.user_constraints = new_con
            if not par.enabled:
                par.enabled = True

    def enforce_sym(self):
        """
        Enforce symmetry constraints on to a Lattice
        """
        # SG system
        crys_system = self.spacegroup.crystal_system
        self.clear_sym()
        trig_test = crys_system == "trigonal" and (
                self.spacegroup.setting_str.endswith("H") or
                self.spacegroup.int_number in [143, 144, 145, 147, 149, 150, 151, 152,
                                               153, 154, 156, 157, 158, 159, 162, 163,
                                               164, 165])

        # Go through the cell systems
        if crys_system == "cubic":
            self.length_a.user_constraints['sg_1'] = ObjConstraint(self.length_b, '', self.length_a)
            self.length_a.user_constraints['sg_1']()
            self.length_a.user_constraints['sg_2'] = ObjConstraint(self.length_c, '', self.length_a)
            self.length_a.user_constraints['sg_2']()
            self.angle_alpha = 90
            self.angle_alpha.enabled = False
            self.angle_beta = 90
            self.angle_beta.enabled = False
            self.angle_gamma = 90
            self.angle_gamma.enabled = False
        elif crys_system == "hexagonal" or trig_test:
            self.length_a.user_constraints["sg_1"] = ObjConstraint(
                self.length_b, "", self.length_a
            )
            self.length_a.user_constraints["sg_1"]()
            self.angle_alpha = 90
            self.angle_alpha.enabled = False
            self.angle_beta = 90
            self.angle_beta.enabled = False
            self.angle_gamma = 120
            self.angle_gamma.enabled = False
        elif crys_system == "trigonal" and not trig_test:
            self.length_a.user_constraints["sg_1"] = ObjConstraint(
                self.length_b, "", self.length_a
            )
            self.length_a.user_constraints["sg_1"]()
            self.length_a.user_constraints["sg_2"] = ObjConstraint(
                self.length_c, "", self.length_a
            )
            self.length_a.user_constraints["sg_2"]()
            self.angle_alpha.user_constraints["sg_1"] = ObjConstraint(
                self.angle_beta, "", self.angle_alpha
            )
            self.angle_alpha.user_constraints["sg_1"]()
            self.angle_alpha.user_constraints["sg_2"] = ObjConstraint(
                self.angle_gamma, "", self.angle_alpha
            )
            self.angle_alpha.user_constraints["sg_2"]()
        elif crys_system == "tetragonal":
            self.length_a.user_constraints['sg_1'] = ObjConstraint(self.length_b, '', self.length_a)
            self.length_a.user_constraints['sg_1']()
            self.angle_alpha = 90
            self.angle_alpha.enabled = False
            self.angle_beta = 90
            self.angle_beta.enabled = False
            self.angle_gamma = 90
            self.angle_gamma.enabled = False
        elif crys_system == "orthorhombic":
            self.angle_alpha = 90
            self.angle_alpha.enabled = False
            self.angle_beta = 90
            self.angle_beta.enabled = False
            self.angle_gamma = 90
            self.angle_gamma.enabled = False
        elif crys_system == "monoclinic":
            self.angle_alpha = 90
            self.angle_alpha.enabled = False
            self.angle_gamma = 90
            self.angle_gamma.enabled = False
        elif crys_system == 'triclinic':
            return
        else:
            raise TypeError('The current crystal system is unknown so symmetry cannot be enforced')

    def to_cell(self) -> L:
        """
        Convert the PeriodicLattice to a standard Lattice. i.e. drop spacegroup. The return is a copy

        :return: Lattice with spacegroup information dropped.
        :rtype: Lattice
        """
        return Lattice(*self.lengths, *self.angles, interface=None)

    def __copy__(self):
        """
        Returns a deep copy of the Periodic-Lattice. Note that the spacegroup parameter is also a copy!!

        :return: Deep copy of self
        :rtype: PeriodicLattice
        """
        return self.__class__(*self.lengths, *self.angles, self.spacegroup.hermann_mauguin, interface=self.interface)

    def __new_SG_setter(self, obj, value):
        self.clear_sym()
        self.__previous_SG_setter(obj, value)
        self.enforce_sym()


def get_integer_index(miller_index: Sequence[float], round_dp: int = 4, verbose: bool = True) -> Tuple[int, int, int]:
    """
    Attempt to convert a vector of floats to whole numbers.
    Args:
        miller_index (list of float): A list miller indexes.
        round_dp (int, optional): The number of decimal places to round the
            miller index to.
        verbose (bool, optional): Whether to print warnings.
    Returns:
        (tuple): The Miller index.
    """
    mi = np.asarray(miller_index)
    # deal with the case we have small irregular floats
    # that are all equal or factors of each other
    mi /= min([m for m in mi if m != 0])
    mi /= np.max(np.abs(mi))

    # deal with the case we have nice fractions
    md = [Fraction(n).limit_denominator(12).denominator for n in mi]
    mi *= reduce(lambda x, y: x * y, md)
    int_miller_index = np.int_(np.round(mi, 1))
    mi /= np.abs(reduce(math.gcd, int_miller_index))

    # round to a reasonable precision
    mi = np.array([round(h, round_dp) for h in mi])

    # need to recalculate this after rounding as values may have changed
    int_miller_index = np.int_(np.round(mi, 1))
    if np.any(np.abs(mi - int_miller_index) > 1e-6) and verbose:
        warnings.warn("Non-integer encountered in Miller index")
    else:
        mi = int_miller_index

    # minimise the number of negative indexes
    mi += 0  # converts -0 to 0

    def n_minus(index):
        return len([h for h in index if h < 0])

    if n_minus(mi) > n_minus(mi * -1):
        mi *= -1

    # if only one index is negative, make sure it is the smallest
    # e.g. (-2 1 0) -> (2 -1 0)
    if (
            sum(mi != 0) == 2
            and n_minus(mi) == 1
            and abs(min(mi)) > max(mi)
    ):
        mi *= -1

    return tuple(mi)  # type: ignore


def get_points_in_spheres(all_coords: np.ndarray, center_coords: np.ndarray, r: float,
                          pbc: Union[bool, List[bool]] = True, numerical_tol: float = 1e-8,
                          lattice: Lattice = None, return_fcoords: bool = False,
                          ) -> List[List[Tuple[np.ndarray, float, int, np.ndarray]]]:
    """
    For each point in `center_coords`, get all the neighboring points in `all_coords` that are within the
    cutoff radius `r`.
    Args:
        all_coords: (list of cartesian coordinates) all available points
        center_coords: (list of cartesian coordinates) all centering points
        r: (float) cutoff radius
        pbc: (bool or a list of bool) whether to set periodic boundaries
        numerical_tol: (float) numerical tolerance
        lattice: (Lattice) lattice to consider when PBC is enabled
        return_fcoords: (bool) whether to return fractional coords when pbc is set.
    Returns:
        List[List[Tuple[coords, distance, index, image]]]
    """
    if isinstance(pbc, bool):
        pbc = [pbc] * 3
    pbc = np.array(pbc, dtype=bool)
    if return_fcoords and lattice is None:
        raise ValueError("Lattice needs to be supplied to compute fractional coordinates")
    center_coords_min = np.min(center_coords, axis=0)
    center_coords_max = np.max(center_coords, axis=0)
    # The lower bound of all considered atom coords
    global_min = center_coords_min - r - numerical_tol
    global_max = center_coords_max + r + numerical_tol
    if np.any(pbc):
        if lattice is None:
            raise ValueError("Lattice needs to be supplied when considering periodic boundary")
        recp_len = np.array(lattice.reciprocal_lattice.lengths)
        maxr = np.ceil((r + 0.15) * recp_len / (2 * math.pi))
        frac_coords = lattice.get_fractional_coords(center_coords)
        nmin_temp = np.floor(np.min(frac_coords, axis=0)) - maxr
        nmax_temp = np.ceil(np.max(frac_coords, axis=0)) + maxr
        nmin = np.zeros_like(nmin_temp)
        nmin[pbc] = nmin_temp[pbc]
        nmax = np.ones_like(nmax_temp)
        nmax[pbc] = nmax_temp[pbc]
        all_ranges = [np.arange(x, y, dtype='int64') for x, y in zip(nmin, nmax)]
        matrix = lattice.matrix
        # temporarily hold the fractional coordinates
        image_offsets = lattice.get_fractional_coords(all_coords)
        all_fcoords = []
        # only wrap periodic boundary
        for k in range(3):
            if pbc[k]:  # type: ignore
                all_fcoords.append(np.mod(image_offsets[:, k:k + 1], 1))
            else:
                all_fcoords.append(image_offsets[:, k:k + 1])
        all_fcoords = np.concatenate(all_fcoords, axis=1)
        image_offsets = image_offsets - all_fcoords
        coords_in_cell = np.dot(all_fcoords, matrix)
        # Filter out those beyond max range
        valid_coords = []
        valid_images = []
        valid_indices = []
        for image in itertools.product(*all_ranges):
            coords = np.dot(image, matrix) + coords_in_cell
            valid_index_bool = np.all(np.bitwise_and(coords > global_min[None, :], coords < global_max[None, :]),
                                      axis=1)
            ind = np.arange(len(all_coords))
            if np.any(valid_index_bool):
                valid_coords.append(coords[valid_index_bool])
                valid_images.append(np.tile(image, [np.sum(valid_index_bool), 1]) - image_offsets[valid_index_bool])
                valid_indices.extend([k for k in ind if valid_index_bool[k]])
        if len(valid_coords) < 1:
            return [[]] * len(center_coords)
        valid_coords = np.concatenate(valid_coords, axis=0)
        valid_images = np.concatenate(valid_images, axis=0)

    else:
        valid_coords = all_coords
        valid_images = [[0, 0, 0]] * len(valid_coords)
        valid_indices = np.arange(len(valid_coords))

    # Divide the valid 3D space into cubes and compute the cube ids
    all_cube_index = _compute_cube_index(valid_coords, global_min, r)
    nx, ny, nz = _compute_cube_index(global_max, global_min, r) + 1
    all_cube_index = _three_to_one(all_cube_index, ny, nz)
    site_cube_index = _three_to_one(_compute_cube_index(center_coords, global_min, r), ny, nz)
    # create cube index to coordinates, images, and indices map
    cube_to_coords = collections.defaultdict(list)  # type: Dict[int, List]
    cube_to_images = collections.defaultdict(list)  # type: Dict[int, List]
    cube_to_indices = collections.defaultdict(list)  # type: Dict[int, List]
    for i, j, k, l in zip(all_cube_index.ravel(), valid_coords,
                          valid_images, valid_indices):
        cube_to_coords[i].append(j)
        cube_to_images[i].append(k)
        cube_to_indices[i].append(l)

    # find all neighboring cubes for each atom in the lattice cell
    site_neighbors = find_neighbors(site_cube_index, nx, ny, nz)
    neighbors = []  # type: List[List[Tuple[np.ndarray, float, int, np.ndarray]]]

    for i, j in zip(center_coords, site_neighbors):
        l1 = np.array(_three_to_one(j, ny, nz), dtype=int).ravel()
        # use the cube index map to find the all the neighboring
        # coords, images, and indices
        ks = [k for k in l1 if k in cube_to_coords]
        if not ks:
            neighbors.append([])
            continue
        nn_coords = np.concatenate([cube_to_coords[k] for k in ks], axis=0)
        nn_images = itertools.chain(*[cube_to_images[k] for k in ks])
        nn_indices = itertools.chain(*[cube_to_indices[k] for k in ks])
        dist = np.linalg.norm(nn_coords - i[None, :], axis=1)
        nns: List[Tuple[np.ndarray, float, int, np.ndarray]] = []
        for coord, index, image, d in zip(nn_coords, nn_indices, nn_images, dist):
            # filtering out all sites that are beyond the cutoff
            # Here there is no filtering of overlapping sites
            if d < r + numerical_tol:
                if return_fcoords and (lattice is not None):
                    coord = np.round(lattice.get_fractional_coords(coord), 10)
                nn = (coord, float(d), int(index), image)
                nns.append(nn)
        neighbors.append(nns)
    return neighbors


# The following internal methods are used in the get_points_in_sphere method.
def _compute_cube_index(coords: np.ndarray, global_min: float, radius: float
                        ) -> np.ndarray:
    """
    Compute the cube index from coordinates
    Args:
        coords: (nx3 array) atom coordinates
        global_min: (float) lower boundary of coordinates
        radius: (float) cutoff radius
    Returns: (nx3 array) int indices
    """
    return np.array(np.floor((coords - global_min) / radius), dtype=int)


def _one_to_three(label1d: np.ndarray, ny: int, nz: int) -> np.ndarray:
    """
    Convert a 1D index array to 3D index array
    Args:
        label1d: (array) 1D index array
        ny: (int) number of cells in y direction
        nz: (int) number of cells in z direction
    Returns: (nx3) int array of index
    """
    last = np.mod(label1d, nz)
    second = np.mod((label1d - last) / nz, ny)
    first = (label1d - last - second * nz) / (ny * nz)
    return np.concatenate([first, second, last], axis=1)


def _three_to_one(label3d: np.ndarray, ny: int, nz: int) -> np.ndarray:
    """
    The reverse of _one_to_three
    """
    return np.array(label3d[:, 0] * ny * nz +
                    label3d[:, 1] * nz + label3d[:, 2]).reshape((-1, 1))


def find_neighbors(label: np.ndarray, nx: int, ny: int, nz: int
                   ) -> List[np.ndarray]:
    """
    Given a cube index, find the neighbor cube indices
    Args:
        label: (array) (n,) or (n x 3) indice array
        nx: (int) number of cells in y direction
        ny: (int) number of cells in y direction
        nz: (int) number of cells in z direction
    Returns: neighbor cell indices
    """

    array = [[-1, 0, 1]] * 3
    neighbor_vectors = np.array(list(itertools.product(*array)),
                                dtype=int)
    if np.shape(label)[1] == 1:
        label3d = _one_to_three(label, ny, nz)
    else:
        label3d = label
    all_labels = label3d[:, None, :] - neighbor_vectors[None, :, :]
    filtered_labels = []
    # filter out out-of-bound labels i.e., label < 0
    for labels in all_labels:
        ind = (labels[:, 0] < nx) * (labels[:, 1] < ny) * (labels[:, 2] < nz) * np.all(labels > -1e-5, axis=1)
        filtered_labels.append(labels[ind])
    return filtered_labels



#  SPDX-FileCopyrightText: 2023 easyCrystallography contributors <crystallography@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2022-2023  Contributors to the easyCore project <https://github.com/easyScience/easyCrystallography>

from __future__ import annotations

__author__ = "github.com/wardsimon"
__version__ = "0.1.0"

from typing import Dict, Union, List, ClassVar, Optional, TYPE_CHECKING

from easyCore import np
from easyCore.Objects.ObjectClasses import BaseObj, Parameter, Descriptor
from easyCore.Objects.Groups import BaseCollection

from easyCrystallography.Components.Lattice import Lattice, PeriodicLattice
from easyCrystallography.Components.Site import Site, PeriodicAtoms, Atoms
from easyCrystallography.Components.SpaceGroup import SpaceGroup
from easyCrystallography.io.parser import Parsers

if TYPE_CHECKING:
    from easyCore.Utils.typing import iF


class Phase(BaseObj):
    _SITE_CLASS = Site
    _ATOMS_CLASS = Atoms

    cell = ClassVar[PeriodicLattice]
    _spacegroup = ClassVar[SpaceGroup]
    atoms = ClassVar[PeriodicAtoms]
    scale = ClassVar[Parameter]

    _REDIRECT = {
        'spacegroup': lambda obj: getattr(obj, '_spacegroup'),
    }

    def __init__(
            self,
            name: str,
            spacegroup: Optional[Union[SpaceGroup, str]] = None,
            cell: Optional[Union[Lattice, PeriodicLattice]] = None,
            atoms: Optional[Atoms] = None,
            scale: Optional[Parameter] = None,
            interface: Optional[iF] = None,
            enforce_sym: bool = True,
    ):
        self.name = name
        if spacegroup is None:
            spacegroup = SpaceGroup()
        if cell is None:
            cell = Lattice()
        if isinstance(cell, Lattice):
            cell = PeriodicLattice.from_lattice_and_spacegroup(cell, spacegroup)
        if atoms is None:
            atoms = Atoms("atoms")
        if scale is None:
            scale = Parameter("scale", 1, min=0)

        super(Phase, self).__init__(
            name, cell=cell, _spacegroup=spacegroup, atoms=atoms, scale=scale
        )
        if not enforce_sym:
            self.cell.clear_sym()
        self._enforce_sym = enforce_sym
        self.interface = interface

        self._extent = np.array([1, 1, 1])
        self._centre = np.array([0, 0, 0])
        self.atom_tolerance = 1e-4

    def add_atom(self, *args, **kwargs):
        """
        Add an atom to the crystal
        """
        supplied_atom = False
        for arg in args:
            if issubclass(arg.__class__, self._SITE_CLASS):
                self.atoms.append(arg)
                supplied_atom = True
        if not supplied_atom:
            atom = Site(*args, **kwargs)
            self.atoms.append(atom)

    def remove_atom(self, key):
        del self.atoms[key]

    def all_orbits(
            self, extent=None, magnetic_only: bool = False
    ) -> Dict[str, np.ndarray]:
        """
        Generate all atomic positions from the atom array and symmetry operations over an extent.

        :return:  dictionary with keys of atom labels, containing numpy arrays of unique points in the extent
        (0, 0, 0) -> obj.extent
        :rtype: Dict[str, np.ndarray]
        """

        if extent is None:
            extent = self._extent

        offsets = np.array(
            np.meshgrid(
                range(0, extent[0] + 1),
                range(0, extent[1] + 1),
                range(0, extent[2] + 1),
            )
        ).T.reshape(-1, 3)

        orbits = self.get_orbits(magnetic_only=magnetic_only)
        for orbit_key in orbits.keys():
            orbit = orbits[orbit_key]
            site_positions = (
                    np.apply_along_axis(np.add, 1, offsets, orbit).reshape((-1, 3))
                    - self.center
            )
            orbits[orbit_key] = (
                    site_positions[
                    np.all(site_positions >= -self.atom_tolerance, axis=1)
                    & np.all(site_positions <= extent + self.atom_tolerance, axis=1),
                    :,
                    ]
                    + self.center
            )
        return orbits

    def get_orbits(self, magnetic_only: bool = False) -> Dict[str, np.ndarray]:
        """
        Generate all atomic positions from the atom array and symmetry operations over an extent.

        :return:  dictionary with keys of atom labels, containing numpy arrays of unique points in the extent
        (0, 0, 0) -> obj.extent
        :rtype: Dict[str, np.ndarray]
        """
        atoms = PeriodicAtoms.from_atoms(self.cell, self.atoms)
        orbits = atoms.get_orbits(magnetic_only=magnetic_only)
        return orbits

    @property
    def enforce_sym(self):
        return self._enforce_sym

    @enforce_sym.setter
    def enforce_sym(self, value: bool):
        if value:
            self.cell.enforce_sym()
        else:
            self.cell.clear_sym()

    @property
    def spacegroup(self):
        return self._spacegroup

    def set_spacegroup(self, value):
        if self._enforce_sym:
            self.cell.space_group_HM_name = value
        else:
            self._spacegroup.space_group_HM_name = value

    @property
    def extent(self) -> np.ndarray:
        """
        Get the current extent in unit cells

        :return: current extent in unit cells
        :rtype: np.ndarray
        """
        return self._extent

    @extent.setter
    def extent(self, new_extent: Union[list, np.ndarray]):
        """
        The current extent of in unit cells. Default (1, 1, 1)

        :param new_extent: The new extent in unit cells.
        :type new_extent: Union[list, tuple, np.ndarray]
        """
        if isinstance(new_extent, list):
            new_extent = np.array(new_extent)
        if np.prod(new_extent.shape) != 3:
            raise ValueError
        new_extent = new_extent.reshape((3,))
        self._extent = new_extent

    @property
    def center(self) -> np.ndarray:
        """
        Get the center position

        :return: center position
        :rtype: np.ndarray
        """
        return self._centre

    @center.setter
    def center(self, new_center: Union[list, tuple, np.ndarray]):
        """
        Set the center position. Default (0, 0, 0)

        :param new_center: New center position.
        :type new_center: Union[list, tuple, np.ndarray]
        """

        if isinstance(new_center, list):
            new_center = np.array(new_center)
        if np.prod(new_center.shape) != 3:
            raise ValueError
        new_center = new_center.reshape((3,))
        self._centre = new_center

    def _generate_positions(self, site, extent) -> np.ndarray:
        """
        Generate all orbits for a given fractional position.
        """
        sym_op = self.spacegroup._sg_data.get_orbit
        offsets = np.array(
            np.meshgrid(
                range(0, extent[0] + 1),
                range(0, extent[1] + 1),
                range(0, extent[2] + 1),
            )
        ).T.reshape(-1, 3)
        return np.apply_along_axis(
            np.add, 1, offsets, np.array(sym_op(site.fract_coords))
        ).reshape((-1, 3))

    def all_sites(self, extent=None) -> Dict[str, np.ndarray]:
        """
        Generate all atomic positions from the atom array and symmetry operations over an extent.
        :return:  dictionary with keys of atom labels, containing numpy arrays of unique points in the extent
        (0, 0, 0) -> obj.extent
        :rtype: Dict[str, np.ndarray]
        """
        if self.spacegroup is None:
            return {atom.label: atom.fract_coords for atom in self.atoms}

        if extent is None:
            extent = self._extent

        sites = {}
        for site in self.atoms:
            unique_sites = self._generate_positions(site, extent)
            site_positions = unique_sites - self.center
            sites[site.label.raw_value] = (
                    site_positions[
                    np.all(site_positions >= -self.atom_tolerance, axis=1)
                    & np.all(site_positions <= extent + self.atom_tolerance, axis=1),
                    :,
                    ]
                    + self.center
            )
        return sites

    @property
    def cif(self) -> str:
        s = ''
        cif_str_parser = Parsers('cif_str')
        with cif_str_parser.writer() as r:
            s += r.structure(self)
        return s

    @classmethod
    def from_cif_file(cls, filename):
        s = None
        with Parsers('cif').reader(filename) as r:
            s = r.structure(phase_class=cls)
        return s


class Phases(BaseCollection):
    _SITE_CLASS = Site
    _ATOM_CLASS = Atoms
    _PHASE_CLASS = Phase

    def __init__(self, name: str = "phases", *args, interface: Optional[iF] = None, **kwargs):
        """
        Generate a collection of crystals.

        :param name: Name of the crystals collection
        :type name: str
        :param args: objects to create the crystal
        :type args: *Phase
        """
        if not isinstance(name, str):
            raise AttributeError("Name should be a string!")

        super(Phases, self).__init__(name, *args, **kwargs)
        self.interface = interface

    def __repr__(self) -> str:
        return f"Collection of {len(self)} phases."

    def __getitem__(
            self, idx: Union[int, slice]
    ) -> Union[Parameter, Descriptor, BaseObj, BaseCollection]:
        if isinstance(idx, str) and idx in self.phase_names:
            idx = self.phase_names.index(idx)
        return super(Phases, self).__getitem__(idx)

    def __delitem__(self, key):
        if isinstance(key, str) and key in self.phase_names:
            key = self.phase_names.index(key)
        return super(Phases, self).__delitem__(key)

    def append(self, item: Phase):
        if not isinstance(item, Phase):
            raise TypeError("Item must be a Phase")
        if item.name in self.phase_names:
            raise AttributeError(f"A phase of name {item.name} already exists.")
        super(Phases, self).append(item)

    @property
    def phase_names(self) -> List[str]:
        return [phase.name for phase in self]

    @property
    def cif(self) -> str:
        s = ''
        cif_str_parser = Parsers('cif_str')
        with cif_str_parser.writer() as r:
            s += r.structures(self)
        return s

    @classmethod
    def from_cif_file(cls, filename):
        s = None
        with Parsers('cif').reader(filename) as r:
            s = r.structures(phases_class=cls, phase_class=cls._PHASE_CLASS)
        return s



from __future__ import annotations

__author__ = 'github.com/wardsimon'
__version__ = '0.0.1'

#  SPDX-FileCopyrightText: 2023 easyCrystallography contributors <crystallography@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2022-2023  Contributors to the easyCore project <https://github.com/easyScience/easyCrystallography>

from typing import List, NoReturn, TYPE_CHECKING, ClassVar, Tuple

from .template import CIF_Template, gemmi

if TYPE_CHECKING:
    from easyCore.Utils.typing import B


class generic(CIF_Template):


    def from_cif_block(self, block: gemmi.cif.Block) -> B:
        kwargs = {}
        errors = {}
        is_fixed = {}
        for item in self._CIF_CONVERSIONS:
            value = block.find_pair_item(self._CIF_SECTION_NAME + item[1])
            V, E, F = self.string_to_variable(value.pair[1])
            if E:
                errors[item[0]] = E
            if F is not None and not F:
                is_fixed[item[0]] = F
            kwargs[item[0]] = V
        obj = self._CIF_CLASS(**kwargs)
        for error in errors.keys():
            setattr(getattr(obj, error), 'error', errors[error])
        for atr in is_fixed.keys():
            setattr(getattr(obj, atr), 'fixed', is_fixed[atr])
        return obj

    def add_to_cif_block(self, obj: B, block: gemmi.cif.Block) -> NoReturn:
        for item in self._CIF_CONVERSIONS:
            value = getattr(obj, item[0])
            block.set_pair(self._CIF_SECTION_NAME + item[1], self.variable_to_string(value))

    def from_cif_string(self, cif_string: str) -> List[B]:

        if "data_" not in cif_string:
            cif_string = "data_temp\n" + cif_string

        cif_blocks = gemmi.cif.read_string(cif_string)
        objs = []
        for block in cif_blocks:
            objs.append(self.from_cif_block(block))
        return objs



#  SPDX-FileCopyrightText: 2023 easyCrystallography contributors <crystallography@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2022-2023  Contributors to the easyCore project <https://github.com/easyScience/easyCrystallography>


__author__ = "github.com/wardsimon"
__version__ = "0.1.0"

from typing import List, Dict, Union, Any
from copy import deepcopy
from easyCore.Objects.ObjectClasses import Descriptor
from easyCore.Utils.classTools import addProp, removeProp
import periodictable as pt
import re

_SPECIE_DETAILS = {
    "type_symbol": {
        "description": "A code to identify the atom species occupying this site.",
        "url":         "https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Iatom_site_type_symbol.html",
    },
}

_REDIRECT = deepcopy(Descriptor._REDIRECT)
_REDIRECT['specie'] = lambda obj: obj._raw_data['str']


class Specie(Descriptor):

    _REDIRECT = _REDIRECT

    def __init__(self, specie: str = "H", **kwargs):
        if "value" in kwargs.keys():
            specie = kwargs.pop("value")

        self._raw_data: Dict[str, Union[str, int]] = {}
        self._props: Dict[str, Any] = {}

        self._reset_data()
        super(Specie, self).__init__("specie", self.__gen_data(specie), **_SPECIE_DETAILS["type_symbol"])

        # Monkey patch the unit and the value to take into account the new type situation
        self.__previous_set = self.__class__.value.fset

        addProp(
            self,
            "value",
            fget=self.__class__.value.fget,
            fset=lambda obj, val: self.__previous_set(obj, obj.__gen_data(val)),
            fdel=self.__class__.value.fdel,
        )

    def _reset_data(self):
        self._raw_data = dict.fromkeys(["str", "observed", "element", "isotope", "oxi_state", "spin"])
        self._props = {}

    def __gen_data(self, value_str: str):
        s = re.search(r"([0-9.]*)([A-Z][a-z]*)([0-9.]*)([+\-]*)", value_str)
        # group(1) = Isotope
        # group(2) = Element
        # group(3) = Oxi state
        # group(4) = Oxi +/2

        if s is None:
            raise ValueError("Invalid specie string")
        element_str = s.group(2)
        element = getattr(pt, element_str, None)
        if element is None:
            raise ValueError(f"Element ({s.group(2)}) not found in periodictable")

        if self._props is not None:
            for k in self._props.keys():
                removeProp(self, k)

        self._reset_data()
        self._raw_data["str"] = value_str
        self._raw_data["element"] = element
        self._raw_data["observed"] = element

        props = {s: getattr(self._raw_data["observed"], s)
                 for s in self._raw_data["observed"].__dir__()
                 if not s.startswith('_') and hasattr(self._raw_data["observed"], s)
                 }
        isotope_str = s.group(1)
        self._raw_data["isotope"] = None
        if isotope_str:
            self._raw_data["isotope"] = int(isotope_str)
            self._raw_data["observed"] = pt.core.Isotope(self._raw_data["observed"], self._raw_data["isotope"])
            props.update(
                {s: getattr(self._raw_data["observed"], s)
                 for s in self._raw_data["observed"].__dir__()
                 if not s.startswith('_') and hasattr(self._raw_data["observed"], s)
                 }
            )
        oxi_state_str = s.group(3)
        oxi_pm_str = s.group(4)
        self._raw_data["oxi_state"] = None
        if oxi_state_str:
            self._raw_data["oxi_state"] = int(oxi_pm_str + oxi_state_str)
            self._raw_data["observed"] = pt.core.Ion(self._raw_data["observed"], self._raw_data["oxi_state"])
            props.update(
                {s: getattr(self._raw_data["observed"], s)
                 for s in self._raw_data["observed"].__dir__()
                 if not s.startswith('_') and hasattr(self._raw_data["observed"], s)
                 }
            )
        props["common_name"] = props.pop("name")
        self._props = props
        for k, v in props.items():
            addProp(self, k, fget=self.__getter_periodic(k))
        rep = f"{self._raw_data['element']}"
        if self.is_ion:
            rep += f"{self._raw_data['oxi_state']}"
            if self._raw_data["oxi_state"] > 0:
                rep += "+"
        return rep

    def __repr__(self) -> str:
        rep = f"<{self.__class__.__name__} \'{self.name}\': "
        if self.is_isotope:
            rep += f"{self._raw_data['isotope']}"
        rep += f"{self._raw_data['element']}"
        if self.is_ion:
            if self._raw_data["oxi_state"] > 0:
                rep += f"{self._raw_data['oxi_state']}+"
            else:
                rep += f"{self._raw_data['oxi_state']}-"
        rep += ">"
        return rep

    def __str__(self) -> str:
        rep = ''
        if self.is_isotope:
            rep += f"{self._raw_data['isotope']}"
        rep += f"{self._raw_data['element']}"
        if self.is_ion:
            if self._raw_data["oxi_state"] > 0:
                rep += f"{self._raw_data['oxi_state']}+"
            else:
                rep += f"{str(self._raw_data['oxi_state'])[1:]}-"
        return rep

    @staticmethod
    def __getter_periodic(key: str):
        def getter(obj):
            return obj._props.get(key)

        return getter

    @property
    def is_ion(self):
        return self._raw_data["oxi_state"] is not None

    @property
    def is_isotope(self):
        return self._raw_data["isotope"] is not None


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['easyCrystallography',
 'easyCrystallography.Components',
 'easyCrystallography.Elements',
 'easyCrystallography.Structures',
 'easyCrystallography.Symmetry',
 'easyCrystallography.io',
 'easyCrystallography.io.cif']

package_data = \
{'': ['*'], 'easyCrystallography': ['Databases/*']}

install_requires = \
['easysciencecore>=0.3.0,<0.4.0',
 'gemmi>=0.5.8,<0.6.0',
 'periodictable>=1.6.1,<2.0.0']

extras_require = \
{'docs': ['doc8>=0.11.0,<0.12.0',
          'readme-renderer>=35,<38',
          'Sphinx>=4.0.2,<6.0.0',
          'sphinx-rtd-theme>=1.0.0,<2.0.0',
          'sphinx-autodoc-typehints>=1.12.0,<2.0.0',
          'sphinx-gallery>=0.10,<0.12'],
 'test': ['pytest>=7.0.0,<8.0.0',
          'pytest-cov>=3,<5',
          'codecov>=2.1.11,<3.0.0',
          'flake8>=5.0,<6.0']}

setup_kwargs = {
    'name': 'easycrystallography',
    'version': '0.3.0',
    'description': 'Crystallography in easyScience',
    'long_description': '# [![License][50]][51] [![Release][32]][33] [![Downloads][70]][71] [![CI Build][20]][21] \n\n[![CodeFactor][83]][84] [![Codecov][85]][86]\n[![Lines of code][81]](<>) [![Total lines][80]](<>) [![Files][82]](<>)\n\n<img height="80"><img src="https://raw.githubusercontent.com/easyScience/easyCrystallography/develop/resources/images/ec_logo.svg" height="65">\n\n**easyCrystallography** is a library that can be used for the generation and manipulation of crystal structures.\n\n## Install\n\n**easyCrystallography** can be downloaded using pip:\n\n```pip install easyCrystallography```\n\nOr direct from the repository:\n\n```pip install https://github.com/easyScience/easyCrystallography```\n\n## Test\n\nAfter installation, launch the test suite:\n\n```python -m pytest```\n\n[//]: # (## Documentation)\n\n[//]: # ()\n[//]: # (Documentation can be found at:)\n\n[//]: # ()\n[//]: # ([https://easyScience.github.io/easyCore]&#40;https://easyScience.github.io/easyCore&#41;)\n\n## Contributing\nWe absolutely welcome contributions. **easyCore** is maintained by the ESS and on a volunteer basis and thus we need to foster a community that can support user questions and develop new features to make this software a useful tool for all users while encouraging every member of the community to share their ideas.\n\n## License\nWhile **easyCore** is under the BSD-3 license\n\n<!---CI Build Status--->\n\n[20]: https://github.com/easyScience/easyCrystallography/workflows/CI%20using%20pip/badge.svg\n\n[21]: https://github.com/easyScience/easyCrystallography/actions\n\n\n<!---Release--->\n\n[32]: https://img.shields.io/pypi/v/easysciencecore?color=green\n\n[33]: https://pypi.org/project/easyCrystallography\n\n\n<!---License--->\n\n[50]: https://img.shields.io/github/license/easyScience/easyCrystallography.svg\n\n[51]: https://github.com/easyScience/easyCrystallography/blob/master/LICENSE.md\n\n\n<!---Downloads--->\n\n[70]: https://img.shields.io/pypi/dm/easyCrystallography.svg\n\n[71]: https://pypi.org/project/easyCrystallography\n\n<!---Code statistics--->\n\n[80]: https://tokei.rs/b1/github/easyScience/easyCrystallography\n\n[81]: https://tokei.rs/b1/github/easyScience/easyCrystallography?category=code\n\n[82]: https://tokei.rs/b1/github/easyScience/easyCrystallography?category=files\n\n[83]: https://www.codefactor.io/repository/github/easyscience/easyCrystallography/badge\n\n[84]: https://www.codefactor.io/repository/github/easyscience/easyCrystallography\n\n[85]: https://img.shields.io/codecov/c/github/easyScience/easyCrystallography?color=green\n\n[86]: https://app.codecov.io/gh/easyScience/easyCrystallography/\n',
    'author': 'Simon Ward',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/easyScience/easyCrystallography',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<=3.12',
}


setup(**setup_kwargs)



from __future__ import annotations
#  SPDX-FileCopyrightText: 2023 easyCrystallography contributors <crystallography@easyscience.software>
#  SPDX-License-Identifier: BSD-3-Clause
#  © 2022-2023  Contributors to the easyCore project <https://github.com/easyScience/easyCrystallography>


__author__ = 'github.com/wardsimon'
__version__ = '0.1.0'

from typing import List, Union, ClassVar, Optional, Type,  TYPE_CHECKING

from easyCore import np
from easyCore.Utils.io.star import StarEntry, StarSection, StarLoop
from easyCore.Objects.ObjectClasses import BaseObj, Descriptor, Parameter
from easyCore.Utils.classTools import addProp, removeProp

if TYPE_CHECKING:
    from easyCore.Utils.typing import iF

_ANIO_DETAILS = {
    'msp_type': {
        'description': "A standard code used to describe the type of atomic displacement parameters used for the site.",
        'url':         'https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Iatom_site_adp_type.html',
        'value':       'Uani'
    },
    'Cani':     {
        'description': 'Isotropic magnetic susceptibility parameter.',
        'url':         'https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Iatom_site_aniso_U_.html',
        'value':       0.0,
        'units':       'T^-1',
        'fixed':       True,
    },
    'Ciso':     {
        'description': 'Isotropic magnetic susceptibility parameter, or equivalent isotropic magnetic susceptibility '
                       'parameter, C(equiv), in inverted teslas, calculated from anisotropic susceptibility '
                       'components.',
        'value':       0.0,
        'max':         np.inf,
        'units':       'T^-1',
        'fixed':       True,
    },
}


class MSPBase(BaseObj):

    def __init__(self, *args, **kwargs):
        super(MSPBase, self).__init__(*args, **kwargs)

    @property
    def matrix(self) -> np.ndarray:
        matrix = np.zeros([3, 3])
        pars = self.get_parameters()
        if len(pars) == 1:
            np.fill_diagonal(matrix, pars[0].raw_value)
        elif len(pars) == 6:
            matrix[0, 0] = pars[0].raw_value
            matrix[0, 1] = pars[1].raw_value
            matrix[0, 2] = pars[2].raw_value
            matrix[1, 1] = pars[3].raw_value
            matrix[1, 2] = pars[4].raw_value
            matrix[2, 2] = pars[5].raw_value
        return matrix


class Cani(MSPBase):

    chi_11: ClassVar[Parameter]
    chi_12: ClassVar[Parameter]
    chi_13: ClassVar[Parameter]
    chi_22: ClassVar[Parameter]
    chi_23: ClassVar[Parameter]
    chi_33: ClassVar[Parameter]

    def __init__(self,
                 chi_11: Optional[Union[Parameter, float]] = None,
                 chi_12: Optional[Union[Parameter, float]] = None,
                 chi_13: Optional[Union[Parameter, float]] = None,
                 chi_22: Optional[Union[Parameter, float]] = None,
                 chi_23: Optional[Union[Parameter, float]] = None,
                 chi_33: Optional[Union[Parameter, float]] = None,
                 msp_values: Optional[Type[MSPBase]] = None,
                 interface=None):

        super(Cani, self).__init__('Cani',
                                   chi_11=Parameter('chi_11', **_ANIO_DETAILS['Cani']),
                                   chi_12=Parameter('chi_12', **_ANIO_DETAILS['Cani']),
                                   chi_13=Parameter('chi_13', **_ANIO_DETAILS['Cani']),
                                   chi_22=Parameter('chi_22', **_ANIO_DETAILS['Cani']),
                                   chi_23=Parameter('chi_23', **_ANIO_DETAILS['Cani']),
                                   chi_33=Parameter('chi_33', **_ANIO_DETAILS['Cani']),
                                   )
        if chi_11 is not None:
            self.chi_11 = chi_11
        if chi_12 is not None:
            self.chi_12 = chi_12
        if chi_13 is not None:
            self.chi_13 = chi_13
        if chi_22 is not None:
            self.chi_22 = chi_22
        if chi_23 is not None:
            self.chi_23 = chi_23
        if chi_33 is not None:
            self.chi_33 = chi_33
        if msp_values is not None:
            self.chi_11 = msp_values.chi_11
            self.chi_12 = msp_values.chi_12
            self.chi_13 = msp_values.chi_13
            self.chi_22 = msp_values.chi_22
            self.chi_23 = msp_values.chi_23
            self.chi_33 = msp_values.chi_33
        self.interface = interface

class Ciso(MSPBase):
    chi: ClassVar[Parameter]

    def __init__(self, chi: Optional[Union[Parameter, float]] = None,
                 msp_values: Optional[Type[MSPBase]] = None,
                 interface: Optional[iF] = None):
        super(Ciso, self).__init__('Ciso',
                                   chi=Parameter('chi', **_ANIO_DETAILS['Ciso']))
        if chi is not None:
            self.chi = chi
        if msp_values is not None:
            self.chi = msp_values.chi
        self.interface = interface

_AVAILABLE_ISO_TYPES = {
    'Cani': Cani,
    'Ciso': Ciso
}


class MagneticSusceptibility(BaseObj):

    msp_type: ClassVar[Descriptor]
    msp_class: ClassVar[Type[MSPBase]]

    def __init__(self, msp_type: Union[Descriptor, str], interface: Optional = None, **kwargs):
        if isinstance(msp_type, str):
            msp_type = Descriptor('msp_type', msp_type)
        msp_class_name = msp_type.raw_value
        if msp_class_name in _AVAILABLE_ISO_TYPES.keys():
            msp_class = _AVAILABLE_ISO_TYPES[msp_class_name]
            if "msp_class" in kwargs:
                # enable passing chi values directly to constructor
                kwargs['msp_values'] = kwargs['msp_class']
                _ = kwargs.pop("msp_class")
            msp = msp_class(**kwargs, interface=interface)
        else:
            raise AttributeError(f"{msp_class_name} is not a valid magnetic susceptibility type")
        super(MagneticSusceptibility, self).__init__('msp',
                                                     msp_type=msp_type,
                                                     msp_class=msp)
        for par in msp.get_parameters():
            addProp(self, par.name, fget=self.__a_getter(par.name), fset=self.__a_setter(par.name))
        self.interface = interface

    def switch_type(self, msp_string: str, **kwargs):
        if msp_string in _AVAILABLE_ISO_TYPES.keys():
            msp_class = _AVAILABLE_ISO_TYPES[msp_string]
            if kwargs:
                msp_class: MSPBase = msp_class(interface=self.interface, **kwargs)
            else:
                msp_class: MSPBase = msp_class(interface=self.interface)
        else:
            raise AttributeError

        for par in self.msp_class.get_parameters():
            removeProp(self, par.name)
        self.msp_class = msp_class
        self.msp_type = msp_string
        for par in msp_class.get_parameters():
            addProp(self, par.name, fget=self.__a_getter(par.name), fset=self.__a_setter(par.name))

    @property
    def available_types(self) -> List[str]:
        return [name for name in _AVAILABLE_ISO_TYPES.keys()]

    @staticmethod
    def __a_getter(key: str):

        def getter(obj):
            return obj.msp_class._kwargs[key]

        return getter

    @staticmethod
    def __a_setter(key):
        def setter(obj, value):
            obj.msp_class._kwargs[key].value = value
        return setter
